#include <stdlib.h>
#include "yconfig.h"
#include "bindkey.h"
#include <X11/keysym.h>

WMKey gMouseWinMove( XK_Pointer_Button1, kfAlt, "Alt+Pointer_Button1" );
WMKey gMouseWinSize( XK_Pointer_Button3, kfAlt, "Alt+Pointer_Button3" );
WMKey gMouseWinRaise( XK_Pointer_Button1, kfCtrl+kfAlt, "Ctrl+Alt+Pointer_Button1" );
WMKey gMouseWinLower( XK_Pointer_Button1, kfCtrl+kfAlt, "Ctrl+Alt+Pointer_Button1" );
WMKey gMouseWinTabbing( XK_Pointer_Button2, kfShift, "Shift+Pointer_Button2" );

WMKey gKeyWinRaise( XK_F1, kfAlt, "Alt+F1" );
WMKey gKeyWinOccupyAll( XK_F2, kfAlt, "Alt+F2" );
WMKey gKeyWinLower( XK_F3, kfAlt, "Alt+F3" );
WMKey gKeyWinClose( XK_F4, kfAlt, "Alt+F4" );
WMKey gKeyWinRestore( XK_F5, kfAlt, "Alt+F5" );
WMKey gKeyWinNext( XK_F6, kfAlt, "Alt+F6" );
WMKey gKeyWinPrev( XK_F6, kfAlt+kfShift, "Alt+Shift+F6" );
WMKey gKeyWinMove( XK_F7, kfAlt, "Alt+F7" );
WMKey gKeyWinSize( XK_F8, kfAlt, "Alt+F8" );
WMKey gKeyWinMinimize( XK_F9, kfAlt, "Alt+F9" );
WMKey gKeyWinMaximize( XK_F10, kfAlt, "Alt+F10" );
WMKey gKeyWinMaximizeVert( XK_F10, kfAlt+kfShift, "Alt+Shift+F10" );
WMKey gKeyWinMaximizeHoriz( 0, 0, "" );
WMKey gKeyWinFullscreen( XK_F11, kfAlt, "Alt+F11" );
WMKey gKeyWinHide( XK_F12, kfAlt+kfShift, "Alt+Shift+F12" );
WMKey gKeyWinRollup( XK_F12, kfAlt, "Alt+F12" );
WMKey gKeyWinArrangeN( XK_KP_Up, kfCtrl+kfAlt, "Ctrl+Alt+KP_Up" );
WMKey gKeyWinArrangeNE( XK_KP_Prior, kfCtrl+kfAlt, "Ctrl+Alt+KP_Prior" );
WMKey gKeyWinArrangeE( XK_KP_Right, kfCtrl+kfAlt, "Ctrl+Alt+KP_Right" );
WMKey gKeyWinArrangeSE( XK_KP_Next, kfCtrl+kfAlt, "Ctrl+Alt+KP_Next" );
WMKey gKeyWinArrangeS( XK_KP_Down, kfCtrl+kfAlt, "Ctrl+Alt+KP_Down" );
WMKey gKeyWinArrangeSW( XK_KP_End, kfCtrl+kfAlt, "Ctrl+Alt+KP_End" );
WMKey gKeyWinArrangeW( XK_KP_Left, kfCtrl+kfAlt, "Ctrl+Alt+KP_Left" );
WMKey gKeyWinArrangeNW( XK_KP_Home, kfCtrl+kfAlt, "Ctrl+Alt+KP_Home" );
WMKey gKeyWinArrangeC( XK_KP_Begin, kfCtrl+kfAlt, "Ctrl+Alt+KP_Begin" );
WMKey gKeyWinTileLeft( 0, 0, "" );
WMKey gKeyWinTileRight( 0, 0, "" );
WMKey gKeyWinTileTop( 0, 0, "" );
WMKey gKeyWinTileBottom( 0, 0, "" );
WMKey gKeyWinTileTopLeft( 0, 0, "" );
WMKey gKeyWinTileTopRight( 0, 0, "" );
WMKey gKeyWinTileBottomLeft( 0, 0, "" );
WMKey gKeyWinTileBottomRight( 0, 0, "" );
WMKey gKeyWinTileCenter( 0, 0, "" );
WMKey gKeyWinSmartPlace( XK_KP_Begin, kfCtrl+kfAlt+kfShift, "Ctrl+Alt+Shift+KP_Begin" );
WMKey gKeySysSwitchNext( XK_Tab, kfAlt, "Alt+Tab" );
WMKey gKeySysSwitchLast( XK_Tab, kfAlt+kfShift, "Alt+Shift+Tab" );
WMKey gKeySysSwitchClass( XK_grave, kfAlt, "Alt+grave" );
WMKey gKeyTaskBarSwitchNext( 0, 0, "" );
WMKey gKeyTaskBarSwitchPrev( 0, 0, "" );
WMKey gKeyTaskBarMoveNext( 0, 0, "" );
WMKey gKeyTaskBarMovePrev( 0, 0, "" );
WMKey gKeySysWinNext( XK_Escape, kfAlt, "Alt+Esc" );
WMKey gKeySysWinPrev( XK_Escape, kfAlt+kfShift, "Alt+Shift+Esc" );
WMKey gKeySysWinMenu( XK_Escape, kfShift, "Shift+Esc" );
WMKey gKeySysDialog( XK_Delete, kfAlt+kfCtrl, "Alt+Ctrl+Del" );
WMKey gKeySysMenu( XK_Escape, kfCtrl, "Ctrl+Esc" );
WMKey gKeySysWindowList( XK_Escape, kfCtrl+kfAlt, "Alt+Ctrl+Esc" );
WMKey gKeySysWinListMenu( 0, 0, "" );
WMKey gKeySysKeyboardNext( 0, 0, "" );
WMKey gKeySysAddressBar( ' ', kfAlt+kfCtrl, "Alt+Ctrl+space" );
WMKey gKeyWinMenu( ' ', kfAlt, "Alt+space" );
WMKey gKeySysWorkspacePrev( XK_Left, kfAlt+kfCtrl, "Alt+Ctrl+Left" );
WMKey gKeySysWorkspaceNext( XK_Right, kfAlt+kfCtrl, "Alt+Ctrl+Right" );
WMKey gKeySysWorkspaceLast( XK_Down, kfAlt+kfCtrl, "Alt+Ctrl+Down" );
WMKey gKeySysWorkspacePrevTakeWin( XK_Left, kfAlt+kfCtrl+kfShift, "Alt+Ctrl+Shift+Left" );
WMKey gKeySysWorkspaceNextTakeWin( XK_Right, kfAlt+kfCtrl+kfShift, "Alt+Ctrl+Shift+Right" );
WMKey gKeySysWorkspaceLastTakeWin( XK_Down, kfAlt+kfCtrl+kfShift, "Alt+Ctrl+Shift+Down" );
WMKey gKeySysWorkspace1( '1', kfAlt+kfCtrl, "Alt+Ctrl+1" );
WMKey gKeySysWorkspace2( '2', kfAlt+kfCtrl, "Alt+Ctrl+2" );
WMKey gKeySysWorkspace3( '3', kfAlt+kfCtrl, "Alt+Ctrl+3" );
WMKey gKeySysWorkspace4( '4', kfAlt+kfCtrl, "Alt+Ctrl+4" );
WMKey gKeySysWorkspace5( '5', kfAlt+kfCtrl, "Alt+Ctrl+5" );
WMKey gKeySysWorkspace6( '6', kfAlt+kfCtrl, "Alt+Ctrl+6" );
WMKey gKeySysWorkspace7( '7', kfAlt+kfCtrl, "Alt+Ctrl+7" );
WMKey gKeySysWorkspace8( '8', kfAlt+kfCtrl, "Alt+Ctrl+8" );
WMKey gKeySysWorkspace9( '9', kfAlt+kfCtrl, "Alt+Ctrl+9" );
WMKey gKeySysWorkspace10( '0', kfAlt+kfCtrl, "Alt+Ctrl+0" );
WMKey gKeySysWorkspace11( '-', kfAlt+kfCtrl, "Alt+Ctrl+minus" );
WMKey gKeySysWorkspace12( '=', kfAlt+kfCtrl, "Alt+Ctrl+equal" );
WMKey gKeySysWorkspace1TakeWin( '1', kfAlt+kfCtrl+kfShift, "Alt+Ctrl+Shift+1" );
WMKey gKeySysWorkspace2TakeWin( '2', kfAlt+kfCtrl+kfShift, "Alt+Ctrl+Shift+2" );
WMKey gKeySysWorkspace3TakeWin( '3', kfAlt+kfCtrl+kfShift, "Alt+Ctrl+Shift+3" );
WMKey gKeySysWorkspace4TakeWin( '4', kfAlt+kfCtrl+kfShift, "Alt+Ctrl+Shift+4" );
WMKey gKeySysWorkspace5TakeWin( '5', kfAlt+kfCtrl+kfShift, "Alt+Ctrl+Shift+5" );
WMKey gKeySysWorkspace6TakeWin( '6', kfAlt+kfCtrl+kfShift, "Alt+Ctrl+Shift+6" );
WMKey gKeySysWorkspace7TakeWin( '7', kfAlt+kfCtrl+kfShift, "Alt+Ctrl+Shift+7" );
WMKey gKeySysWorkspace8TakeWin( '8', kfAlt+kfCtrl+kfShift, "Alt+Ctrl+Shift+8" );
WMKey gKeySysWorkspace9TakeWin( '9', kfAlt+kfCtrl+kfShift, "Alt+Ctrl+Shift+9" );
WMKey gKeySysWorkspace10TakeWin( '0', kfAlt+kfCtrl+kfShift, "Alt+Ctrl+Shift+0" );
WMKey gKeySysWorkspace11TakeWin( '-', kfAlt+kfCtrl+kfShift, "Alt+Ctrl+Shift+minus" );
WMKey gKeySysWorkspace12TakeWin( '=', kfAlt+kfCtrl+kfShift, "Alt+Ctrl+Shift+equal" );
WMKey gKeySysTileVertical( XK_F2, kfAlt+kfShift, "Alt+Shift+F2" );
WMKey gKeySysTileHorizontal( XK_F3, kfAlt+kfShift, "Alt+Shift+F3" );
WMKey gKeySysCascade( XK_F4, kfAlt+kfShift, "Alt+Shift+F4" );
WMKey gKeySysArrange( XK_F5, kfAlt+kfShift, "Alt+Shift+F5" );
WMKey gKeySysArrangeIcons( XK_F8, kfAlt+kfShift, "Alt+Shift+F8" );
WMKey gKeySysMinimizeAll( XK_F9, kfAlt+kfShift, "Alt+Shift+F9" );
WMKey gKeySysHideAll( XK_F11, kfAlt+kfShift, "Alt+Shift+F11" );
WMKey gKeySysUndoArrange( XK_F7, kfAlt+kfShift, "Alt+Shift+F7" );
WMKey gKeySysShowDesktop( 'd', kfAlt+kfCtrl, "Alt+Ctrl+d" );
WMKey gKeySysCollapseTaskBar( 'h', kfAlt+kfCtrl, "Alt+Ctrl+h" );

// vim: set sw=4 ts=4 et:
